# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
# @author Simon Heybrock
from __future__ import annotations

import inspect
import warnings
from collections.abc import Mapping
from functools import lru_cache
from types import MappingProxyType
from typing import Any, Dict, Iterator, List, Optional, Tuple, Union, overload

import numpy as np
import scipp as sc

from ._cache import cached_property
from ._common import to_child_select
from .attrs import Attrs
from .field import Field
from .typing import H5Dataset, H5Group, ScippIndex


def asvariable(obj: Union[Any, sc.Variable]) -> sc.Variable:
    return obj if isinstance(obj, sc.Variable) else sc.scalar(obj, unit=None)


class NexusStructureError(Exception):
    """Invalid or unsupported class and field structure in Nexus."""

    pass


def is_dataset(obj: Union[H5Group, H5Dataset]) -> bool:
    """Return true if the object is an h5py.Dataset or equivalent.

    Use this instead of isinstance(obj, h5py.Dataset) to ensure that code is compatible
    with other h5py-alike interfaces.
    """
    return hasattr(obj, 'shape')


_scipp_dtype = {
    np.dtype('int8'): sc.DType.int32,
    np.dtype('int16'): sc.DType.int32,
    np.dtype('uint8'): sc.DType.int32,
    np.dtype('uint16'): sc.DType.int32,
    np.dtype('uint32'): sc.DType.int32,
    np.dtype('uint64'): sc.DType.int64,
    np.dtype('int32'): sc.DType.int32,
    np.dtype('int64'): sc.DType.int64,
    np.dtype('float32'): sc.DType.float32,
    np.dtype('float64'): sc.DType.float64,
    np.dtype('bool'): sc.DType.bool,
}


def _dtype_fromdataset(dataset: H5Dataset) -> sc.DType:
    return _scipp_dtype.get(dataset.dtype, sc.DType.string)


def _squeezed_field_sizes(dataset: H5Dataset) -> Dict[str, int]:
    if (shape := dataset.shape) == (1,):
        return {}
    return {f'dim_{i}': size for i, size in enumerate(shape)}


class NXobject:
    def _init_field(self, field: Field):
        if field.sizes is None:
            field.sizes = _squeezed_field_sizes(field.dataset)
        field.dtype = _dtype_fromdataset(field.dataset)

    def __init__(self, attrs: Dict[str, Any], children: Dict[str, Union[Field, Group]]):
        """Subclasses should call this in their __init__ method, or ensure that they
        initialize the fields in `children` with the correct sizes and dtypes."""
        self._attrs = attrs
        self._children = children
        for field in children.values():
            if isinstance(field, Field):
                self._init_field(field)

    @property
    def unit(self) -> Union[None, sc.Unit]:
        raise AttributeError(
            f"Group-like {self._attrs.get('NX_class')} has no well-defined unit"
        )

    @cached_property
    def sizes(self) -> Dict[str, int]:
        return sc.DataGroup(self._children).sizes

    def index_child(
        self, child: Union[Field, Group], sel: ScippIndex
    ) -> Union[sc.Variable, sc.DataArray, sc.Dataset, sc.DataGroup]:
        """
        When a Group is indexed, this method is called to index each child.

        The main purpose of this is to translate the Group index to the child index.
        Since the group dimensions (usually given by the signal) may be a superset of
        the child dimensions, we need to translate the group index to a child index.

        The default implementation assumes that the child shape is identical to the
        group shape, for all child dims. Subclasses of NXobject, in particular NXdata,
        override this method to handle bin edges.
        """
        # TODO Could avoid determining sizes if sel is trivial. Merge with
        # NXdata.index_child?
        child_sel = to_child_select(tuple(self.sizes), child.dims, sel)
        return child[child_sel]

    def read_children(self, sel: ScippIndex) -> sc.DataGroup:
        """
        When a Group is indexed, this method is called to read all children.

        The default implementation simply calls index_child on each child and returns
        the result as a DataGroup.

        Subclasses of NXobject, in particular NXevent_data, override this method to
        to implement special logic for reading children with interdependencies, i.e.,
        where reading each child in isolation is not possible.
        """
        return sc.DataGroup(
            {
                name: self.index_child(child, sel)
                for name, child in self._children.items()
            }
        )

    def assemble(
        self, dg: sc.DataGroup
    ) -> Union[sc.DataGroup, sc.DataArray, sc.Dataset]:
        """
        When a Group is indexed, this method is called to assemble the read children
        into the result object.

        The default implementation simply returns the DataGroup.

        Subclasses of NXobject, in particular NXdata, override this method to return
        an object with more semantics such as a DataArray or Dataset.
        """
        return dg


class NXroot(NXobject):
    pass


class Group(Mapping):
    """
    A group in a NeXus file.

    This class is a wrapper around an h5py.Group object. It provides a dict-like
    interface to the children of the group, and provides access to the attributes
    of the group. The children are either Field or Group objects, depending on
    whether the child is a dataset or a group, respectively.
    """

    # The implementation of this class is unfortunately relatively complex:

    # 1. NeXus requires "nonlocal" information for interpreting a field. For example,
    #    NXdata attributes define which fields are the signal, and the names of axes.
    #    A field cannot be read without this information, in particular since we want to
    #    support reading slices, using the Scipp dimension-label syntax.
    # 2. HDF5 or h5py performance is not great, and we want to avoid reading the same
    #    attrs or datasets multiple times. We can therefore not rely on "on-the-fly"
    #    interpretation of the file, but need to cache information. An earlier version
    #    of ScippNexus used such a mechanism without caching, which was very slow.

    def __init__(self, group: H5Group, definitions: Optional[Dict[str, type]] = None):
        self._group = group
        self._definitions = {} if definitions is None else definitions
        self._lazy_children = None
        self._lazy_nexus = None

    @property
    def nx_class(self) -> Optional[type]:
        """The value of the NX_class attribute of the group.

        In case of the subclass NXroot this returns :py:class:`NXroot` even if the attr
        is not actually set. This is to support the majority of all legacy files, which
        do not have this attribute.
        """
        if (nxclass := self.attrs.get('NX_class')) is not None:
            return _nx_class_registry().get(nxclass)
        if self.name == '/':
            return NXroot

    @cached_property
    def attrs(self) -> Dict[str, Any]:
        """The attributes of the group.

        Cannot be used for writing attributes, since they are cached for performance."""
        # Attrs are not read until needed, to avoid reading all attrs for all subgroups.
        # We may expected a per-subgroup overhead of 1 ms for reading attributes, so if
        # all we want is access one subgroup, we may save, e.g., a second for a group
        # with 1000 subgroups (or subfields).
        return MappingProxyType(Attrs(self._group.attrs))

    @property
    def name(self) -> str:
        return self._group.name

    @property
    def unit(self) -> Optional[sc.Unit]:
        return self._nexus.unit

    @property
    def parent(self) -> Group:
        return Group(self._group.parent, definitions=self._definitions)

    @cached_property
    def file(self) -> Group:
        return Group(self._group.file, definitions=self._definitions)

    @property
    def _children(self) -> Dict[str, Union[Field, Group]]:
        """Lazily initialized children of the group."""
        if self._lazy_children is None:
            self._lazy_children = self._read_children()
        return self._lazy_children

    def _read_children(self) -> Dict[str, Union[Field, Group]]:
        def _make_child(obj: Union[H5Dataset, H5Group]) -> Union[Field, Group]:
            if is_dataset(obj):
                return Field(obj, parent=self)
            else:
                return Group(obj, definitions=self._definitions)

        items = {name: _make_child(obj) for name, obj in self._group.items()}
        # In the case of NXevent_data, the `cue_` fields are unusable, since
        # the definition is broken (the cue_index points into the
        # event_time_offset/event_id fields, instead of the
        # event_time_zero/event_index fields). In the case of NXlog they may
        # be some utility if we deal with extremely long time-series that
        # could be leveraged for label-based indexing in the future.
        items = {k: v for k, v in items.items() if not k.startswith('cue_')}
        for suffix in ('_errors', '_error'):
            field_with_errors = [name for name in items if f'{name}{suffix}' in items]
            for name in field_with_errors:
                values = items[name]
                errors = items[f'{name}{suffix}']
                if (
                    isinstance(values, Field)
                    and isinstance(errors, Field)
                    and (values.unit == errors.unit or errors.unit is None)
                    and values.dataset.shape == errors.dataset.shape
                ):
                    values.errors = errors.dataset
                    del items[f'{name}{suffix}']
        return items

    @property
    def _nexus(self) -> NXobject:
        """Instance of the NXobject subclass corresponding to the NX_class attribute.

        This is used to determine dims, unit, and other attributes of the group and its
        children, as well as defining how children will be read and assembled into the
        result object when the group is indexed.

        Lazily initialized since the NXobject subclass init can be costly.
        """
        self._populate_fields()
        return self._lazy_nexus

    def _populate_fields(self) -> None:
        """Populate the fields of the group.

        Fields are not populated until needed, to avoid reading field and group
        properties for all subgroups. However, when any field is read we must in
        general parse all the field and group properties, since for classes such
        as NXdata the properties of any field may indirectly depend on the properties
        of any other field. For example, field attributes may define which fields are
        axes, and dim labels of other fields can be defined by the names of the axes.
        """
        if self._lazy_nexus is not None:
            return
        self._lazy_nexus = self._definitions.get(self.attrs.get('NX_class'), NXobject)(
            attrs=self.attrs, children=self._children
        )

    def __len__(self) -> int:
        return len(self._children)

    def __iter__(self) -> Iterator[str]:
        return self._children.__iter__()

    def _get_children_by_nx_class(
        self, select: Union[type, List[type]]
    ) -> Dict[str, Union[NXobject, Field]]:
        children = {}
        select = tuple(select) if isinstance(select, list) else select
        for key, child in self._children.items():
            nx_class = Field if isinstance(child, Field) else child.nx_class
            if nx_class is not None and issubclass(nx_class, select):
                children[key] = self[key]
        return children

    @overload
    def __getitem__(self, sel: str) -> Union[Group, Field]:
        ...

    @overload
    def __getitem__(
        self, sel: ScippIndex
    ) -> Union[sc.DataArray, sc.DataGroup, sc.Dataset]:
        ...

    @overload
    def __getitem__(self, sel: Union[type, List[type]]) -> Dict[str, NXobject]:
        ...

    def __getitem__(self, sel):
        """
        Get a child group or child dataset, a selection of child groups, or load and
        return the current group.

        Three cases are supported:

        - String name: The child group or child dataset of that name is returned.
        - Class such as ``NXdata`` or ``NXlog``: A dict containing all direct children
          with a matching ``NX_class`` attribute are returned. Also accepts a tuple of
          classes. ``Field`` selects all child fields, i.e., all datasets but not
          groups.
        - Scipp-style index: Load the specified slice of the current group, returning
          a :class:`scipp.DataArray` or :class:`scipp.DataGroup`.

        Parameters
        ----------
        sel:
            Child name, class, or index.

        Returns
        -------
        :
            Field, group, dict of fields, or loaded data.
        """
        if isinstance(sel, str):
            # We cannot get the child directly from the HDF5 group, since we need to
            # create the parent group, to ensure that fields get the correct properties
            # such as sizes and dtype.
            if '/' in sel:
                if sel.startswith('/'):
                    return self.file[sel[1:]]
                else:
                    grp, path = sel.split('/', 1)
                    return self[grp][path]
            child = self._children[sel]
            if isinstance(child, Field):
                self._populate_fields()
            return child

        def isclass(x):
            return inspect.isclass(x) and issubclass(x, (Field, NXobject))

        if isclass(sel) or (
            isinstance(sel, list) and len(sel) and all(isclass(x) for x in sel)
        ):
            return self._get_children_by_nx_class(sel)

        dg = self._nexus.read_children(sel)
        if not dg:
            # Bail out early to avoid fallback warnings. Everything is optional in
            # NeXus so we cannot assume that the group is invalid (in contrast to
            # likely partially incomplete groups that will fail to assemble below).
            return dg
        try:
            dg = self._nexus.assemble(dg)
        except (sc.DimensionError, NexusStructureError) as e:
            self._warn_fallback(e)
        # For a time-dependent transformation in NXtransformations, an NXlog may
        # take the place of the `value` field. In this case, we need to read the
        # properties of the NXlog group to make the actual transformation.
        from .nxtransformations import maybe_transformation

        return maybe_transformation(self, value=dg, sel=sel)

    def _warn_fallback(self, e: Exception) -> None:
        msg = (
            f"Failed to load {self.name} as {type(self._nexus).__name__}: {e} "
            "Falling back to loading HDF5 group children as scipp.DataGroup."
        )
        warnings.warn(msg, stacklevel=2)

    def __setitem__(self, key, value):
        """Set a child group or child dataset.

        Note that due to the caching mechanisms in this class, reading the group
        or its children may not reflect the changes made by this method."""
        if hasattr(value, '__write_to_nexus_group__'):
            group = create_class(self._group, key, nx_class=value.nx_class)
            value.__write_to_nexus_group__(group)
        else:
            create_field(self._group, key, value)

    def create_field(self, key: str, value: sc.Variable) -> H5Dataset:
        """Create a child dataset with given name and value.

        Note that due to the caching mechanisms in this class, reading the group
        or its children may not reflect the changes made by this method.

        Returns
        -------
        :
            The created dataset of the values.
            If errors are written to the file, their dataset is not returned.
        """
        return create_field(self._group, key, value)

    def create_class(self, name: str, class_name: str) -> Group:
        """Create empty HDF5 group with given name and set the NX_class attribute.

        Note that due to the caching mechanisms in this class, reading the group
        or its children may not reflect the changes made by this method.

        Parameters
        ----------
        name:
            Group name.
        nx_class:
            Nexus class, can be a valid string for the NX_class attribute, or a
            subclass of NXobject, such as NXdata or NXlog.
        """
        return Group(
            create_class(self._group, name, class_name), definitions=self._definitions
        )

    @cached_property
    def sizes(self) -> Dict[str, int]:
        return self._nexus.sizes

    @property
    def dims(self) -> Tuple[str, ...]:
        return tuple(self.sizes)

    @property
    def shape(self) -> Tuple[int, ...]:
        return tuple(self.sizes.values())


def _create_field_params_numpy(data: np.ndarray):
    return data, None, {}


def _create_field_params_string(data: sc.Variable):
    return np.array(data.values, dtype=object), None, {}


def _create_field_params_datetime(data: sc.Variable):
    start = sc.epoch(unit=data.unit)
    return (data - start).values, None, {'start': str(start.value)}


def _create_field_params_number(data: sc.Variable):
    errors = sc.stddevs(data).values if data.variances is not None else None
    return data.values, errors, {}


def create_field(
    group: H5Group, name: str, data: Union[np.ndarray, sc.Variable], **kwargs
) -> H5Dataset:
    if not isinstance(data, sc.Variable):
        values, errors, attrs = _create_field_params_numpy(data)
    elif data.dtype == sc.DType.string:
        values, errors, attrs = _create_field_params_string(data)
    elif data.dtype == sc.DType.datetime64:
        values, errors, attrs = _create_field_params_datetime(data)
    else:
        values, errors, attrs = _create_field_params_number(data)

    if isinstance(data, sc.Variable) and data.unit:
        attrs['units'] = str(data.unit)

    values_dataset = group.create_dataset(name, data=values, **kwargs)
    values_dataset.attrs.update(attrs)
    if errors is not None:
        errors_dataset = group.create_dataset(name + '_errors', data=errors, **kwargs)
        errors_dataset.attrs.update(attrs)

    return values_dataset


def create_class(group: H5Group, name: str, nx_class: Union[str, type]) -> H5Group:
    """Create empty HDF5 group with given name and set the NX_class attribute.

    Parameters
    ----------
    name:
        Group name.
    nx_class:
        Nexus class, can be a valid string for the NX_class attribute, or a
        subclass of NXobject, such as NXdata or NXlog.
    """
    group = group.create_group(name)
    attr = nx_class if isinstance(nx_class, (str, bytes)) else nx_class.__name__
    group.attrs['NX_class'] = attr
    return group


@lru_cache()
def _nx_class_registry():
    from . import nexus_classes

    return dict(inspect.getmembers(nexus_classes, inspect.isclass))


base_definitions_dict = {}


def base_definitions() -> Dict[str, type]:
    """Return a dict of all base definitions.

    This is a copy of the base definitions dict, so that it can be modified without
    affecting the original.
    """
    return dict(base_definitions_dict)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
# @author Simon Heybrock
from __future__ import annotations

from typing import List, Optional, Union
from warnings import warn

import numpy as np
import scipp as sc

from .._common import to_child_select
from ..typing import H5Group
from .nxcylindrical_geometry import NXcylindrical_geometry
from .nxobject import Field, NexusStructureError, NXobject, ScippIndex, asarray
from .nxoff_geometry import NXoff_geometry
from .nxtransformations import NXtransformations


class NXdataStrategy:
    """
    Strategy used by :py:class:`scippnexus.NXdata`.

    May be subclassed to customize behavior.
    """

    _error_suffixes = ['_errors', '_error']  # _error is the deprecated suffix

    @staticmethod
    def axes(group):
        """Names of the axes (dimension labels)."""
        return group.attrs.get('axes')

    @staticmethod
    def signal(group):
        """Name of the signal field."""
        if (name := group.attrs.get('signal')) is not None:
            if name in group:
                return name
        # Legacy NXdata defines signal not as group attribute, but attr on dataset
        for name in group.keys():
            # What is the meaning of the attribute value? It is undocumented, we simply
            # ignore it.
            if 'signal' in group._get_child(name).attrs:
                return name
        return None

    @staticmethod
    def signal_errors(group) -> Optional[str]:
        """Name of the field to use for standard-deviations of the signal."""
        name = f'{NXdataStrategy.signal(group)}_errors'
        if name in group:
            return name
        # This is a legacy named, deprecated in the NeXus format.
        if 'errors' in group:
            return 'errors'

    @staticmethod
    def coord_errors(group, name):
        """Name of the field to use for standard-deviations of a coordinate."""
        errors = [f'{name}{suffix}' for suffix in NXdataStrategy._error_suffixes]
        errors = [x for x in errors if x in group]
        if len(errors) == 0:
            return None
        if len(errors) == 2:
            warn(
                f"Found {name}_errors as well as the deprecated "
                f"{name}_error. The latter will be ignored."
            )
        return errors[0]


class NXdata(NXobject):
    def __init__(
        self,
        group: H5Group,
        *,
        definition=None,
        strategy=None,
        signal_override: Union[Field, '_EventField'] = None,  # noqa: F821
        skip: List[str] = None,
    ):
        """
        Parameters
        ----------
        signal_override:
            Field-like to use instead of trying to read signal from the file. This is
            used when there is no signal or to provide a signal computed from
            NXevent_data.
        skip:
            Names of fields to skip when loading coords.
        """
        super().__init__(group, definition=definition, strategy=strategy)
        self._signal_override = signal_override
        self._skip = skip if skip is not None else []

    def _default_strategy(self):
        return NXdataStrategy

    @property
    def shape(self) -> List[int]:
        return self._signal.shape

    def _get_group_dims(self) -> Union[None, List[str]]:
        # Apparently it is not possible to define dim labels unless there are
        # corresponding coords. Special case of '.' entries means "no coord".
        if (axes := self._strategy.axes(self)) is not None:
            return [f'dim_{i}' if a == '.' else a for i, a in enumerate(axes)]
        axes = []
        # Names of axes that have an "axis" attribute serve as dim labels in legacy case
        for name, field in self._group.items():
            if (axis := field.attrs.get('axis')) is not None:
                axes.append((axis, name))
        if axes:
            return [x[1] for x in sorted(axes)]
        return None

    @property
    def dims(self) -> List[str]:
        if (d := self._get_group_dims()) is not None:
            return d
        # Legacy NXdata defines axes not as group attribute, but attr on dataset.
        # This is handled by class Field.
        return self._signal.dims

    @property
    def unit(self) -> Union[sc.Unit, None]:
        return self._signal.unit

    @property
    def _signal_name(self) -> str:
        return self._strategy.signal(self)

    @property
    def _errors_name(self) -> Optional[str]:
        return self._strategy.signal_errors(self)

    @property
    def _signal(self) -> Union[Field, '_EventField', None]:  # noqa: F821
        if self._signal_override is not None:
            return self._signal_override
        if self._signal_name is not None:
            if self._signal_name not in self:
                raise NexusStructureError(
                    f"Signal field '{self._signal_name}' not found in group."
                )
            return self[self._signal_name]
        return None

    def _get_axes(self):
        """Return labels of named axes. Does not include default 'dim_{i}' names."""
        if (axes := self._strategy.axes(self)) is not None:
            # Unlike self.dims we *drop* entries that are '.'
            return [a for a in axes if a != '.']
        elif (signal := self._signal) is not None:
            if (axes := signal.attrs.get('axes')) is not None:
                dims = axes.split(':')
                # The standard says that the axes should be colon-separated, but some
                # files use comma-separated.
                if len(dims) == 1 and self._signal.ndim > 1:
                    dims = tuple(axes.split(','))
                return dims
        return []

    def _guess_dims(self, name: str):
        """Guess dims of non-signal dataset based on shape.

        Does not check for potential bin-edge coord.
        """
        shape = self._get_child(name).shape
        if self.shape == shape:
            return self.dims
        lut = {}
        if self._signal is not None:
            for d, s in zip(self.dims, self.shape):
                if self.shape.count(s) == 1:
                    lut[s] = d
        try:
            dims = [lut[s] for s in shape]
        except KeyError:
            raise NexusStructureError(
                f"Could not determine axis indices for {self.name}/{name}"
            )
        return dims

    def _try_guess_dims(self, name):
        try:
            return self._guess_dims(name)
        except NexusStructureError:
            return None

    def _get_field_dims(self, name: str) -> Union[None, List[str]]:
        # Newly written files should always contain indices attributes, but the
        # standard recommends that readers should also make "best effort" guess
        # since legacy files do not set this attribute.
        if (indices := self.attrs.get(f'{name}_indices')) is not None:
            return list(np.array(self.dims)[np.array(indices).flatten()])
        if (axis := self._get_child(name).attrs.get('axis')) is not None:
            return (self._get_group_dims()[axis - 1],)
        if name in [self._signal_name, self._errors_name]:
            return self._get_group_dims()  # if None, field determines dims itself
        if name in list(self.attrs.get('auxiliary_signals', [])):
            return self._try_guess_dims(name)
        if name in self._get_axes():
            # If there are named axes then items of same name are "dimension
            # coordinates", i.e., have a dim matching their name.
            # However, if the item is not 1-D we need more labels. Try to use labels of
            # signal if dimensionality matches.
            if self._signal_name in self and self._get_child(name).ndim == len(
                self.shape
            ):
                return self[self._signal_name].dims
            return [name]
        return self._try_guess_dims(name)

    def _bin_edge_dim(self, coord: Field) -> Union[None, str]:
        sizes = dict(zip(self.dims, self.shape))
        for dim, size in zip(coord.dims, coord.shape):
            if dim in sizes and sizes[dim] + 1 == size:
                return dim
        return None

    def _dim_of_coord(self, name: str, coord: Field) -> Union[None, str]:
        if len(coord.dims) == 1:
            return coord.dims[0]
        if name in coord.dims and name in self.dims:
            return name
        return self._bin_edge_dim(coord)

    def _should_be_aligned(self, da: sc.DataArray, name: str, coord: Field) -> bool:
        dim_of_coord = self._dim_of_coord(name, coord)
        if dim_of_coord is None:
            return True
        if dim_of_coord not in da.dims:
            return False
        return True

    def _getitem(self, select: ScippIndex) -> sc.DataArray:
        from .nexus_classes import NXgeometry

        signal = self._signal
        if signal is None:
            raise NexusStructureError("No signal field found, cannot load group.")
        signal = signal[select]
        if self._errors_name is not None:
            stddevs = self[self._errors_name][select]
            # According to the standard, errors must have the same shape as the data.
            # This is not the case in all files we observed, is there any harm in
            # attempting a broadcast?
            signal.variances = np.broadcast_to(
                sc.pow(stddevs, sc.scalar(2)).values, shape=signal.shape
            )

        da = (
            signal
            if isinstance(signal, sc.DataArray)
            else sc.DataArray(data=asarray(signal))
        )

        skip = self._skip
        skip += [self._signal_name, self._errors_name]
        skip += list(self.attrs.get('auxiliary_signals', []))
        for name in self:
            if (errors := self._strategy.coord_errors(self, name)) is not None:
                skip += [errors]
        for name in self:
            if name in skip:
                continue
            # It is not entirely clear whether skipping NXtransformations is the right
            # solution. In principle NXobject will load them via the 'depends_on'
            # mechanism, so for valid files this should be sufficient.
            allowed = (
                Field,
                NXtransformations,
                NXcylindrical_geometry,
                NXoff_geometry,
                NXgeometry,
            )
            if not isinstance(self._get_child(name), allowed):
                raise NexusStructureError(
                    "Invalid NXdata: may not contain nested groups"
                )

        for name, field in self[Field].items():
            if name in skip:
                continue
            sel = to_child_select(
                self.dims, field.dims, select, bin_edge_dim=self._bin_edge_dim(field)
            )
            coord: sc.Variable = asarray(self[name][sel])
            if (error_name := self._strategy.coord_errors(self, name)) is not None:
                stddevs = asarray(self[error_name][sel])
                coord.variances = sc.pow(stddevs, sc.scalar(2)).values
            try:
                da.coords[name] = coord
                da.coords.set_aligned(name, self._should_be_aligned(da, name, field))
            except sc.DimensionError as e:
                raise NexusStructureError(
                    f"Field in NXdata incompatible with dims or shape of signal: {e}"
                ) from e

        return da



import h5py
import pytest
import scipp as sc

import scippnexus as snx


@pytest.fixture()
def nxroot():
    """Yield NXroot containing a single NXentry named 'entry'"""
    with h5py.File('dummy.nxs', mode='w', driver="core", backing_store=False) as f:
        root = snx.Group(f, snx.base_definitions())
        root.create_class('entry', snx.NXentry)
        yield root


def test_vertices_loaded_as_vector3(nxroot):
    shape = nxroot['entry'].create_class('shape', snx.NXcylindrical_geometry)
    values = [[1, 2, 3], [4, 5, 6]]
    shape['vertices'] = sc.array(dims=['ignored', 'comp'], values=values, unit='mm')
    loaded = shape[()]
    assert sc.identical(
        loaded['vertices'], sc.vectors(dims=['vertex'], values=values, unit='mm')
    )


def test_field_properties(nxroot):
    shape = nxroot['entry'].create_class('shape', snx.NXcylindrical_geometry)
    values = [[1, 2, 3], [4, 5, 6], [7, 8, 9]]
    shape['vertices'] = sc.array(dims=['ignored', 'comp'], values=values, unit='m')
    shape['cylinders'] = sc.array(
        dims=['ignored', 'index'], values=[[0, 1, 2]], unit=None
    )
    shape['detector_number'] = sc.array(dims=['ignored'], values=[], unit=None)
    loaded = shape[()]
    assert loaded['vertices'].dims == ('vertex',)
    assert loaded['cylinders'].dims == ('cylinder', 'vertex_index')
    assert loaded['cylinders'].unit is None
    assert loaded['detector_number'].dims == ('detector_number',)
    assert loaded['detector_number'].unit is None



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
# @author Simon Heybrock
from __future__ import annotations

from typing import List, Optional, Union

import numpy as np
import scipp as sc
from scipp.scipy import interpolate

from .nxobject import Field, NexusStructureError, NXobject, ScippIndex


class TransformationError(NexusStructureError):
    pass


def make_transformation(obj, /, path) -> Optional[Transformation]:
    if path.startswith('/'):
        return Transformation(obj.file[path])
    elif path != '.':
        return Transformation(obj.parent[path])
    return None  # end of chain


class NXtransformations(NXobject):
    """Group of transformations."""

    def _getitem(self, index: ScippIndex) -> sc.DataGroup:
        return sc.DataGroup(
            {
                name: get_full_transformation_starting_at(
                    Transformation(child), index=index
                )
                for name, child in self.items()
            }
        )


class Transformation:
    def __init__(self, obj: Union[Field, NXobject]):  # could be an NXlog
        self._obj = obj

    @property
    def attrs(self):
        return self._obj.attrs

    @property
    def name(self):
        return self._obj.name

    @property
    def depends_on(self):
        if (path := self.attrs.get('depends_on')) is not None:
            return make_transformation(self._obj, path)
        return None

    @property
    def offset(self):
        if (offset := self.attrs.get('offset')) is None:
            return None
        if (offset_units := self.attrs.get('offset_units')) is None:
            raise TransformationError(
                f"Found {offset=} but no corresponding 'offset_units' "
                f"attribute at {self.name}"
            )
        return sc.spatial.translation(value=offset, unit=offset_units)

    @property
    def vector(self) -> sc.Variable:
        return sc.vector(value=self.attrs.get('vector'))

    def __getitem__(self, select: ScippIndex):
        transformation_type = self.attrs.get('transformation_type')
        # According to private communication with Tobias Richter, NeXus allows 0-D or
        # shape=[1] for single values. It is unclear how and if this could be
        # distinguished from a scan of length 1.
        value = self._obj[select]
        try:
            if isinstance(value, sc.DataGroup):
                raise TransformationError(
                    f"Failed to load transformation at {self.name}."
                )
            t = value * self.vector
            v = t if isinstance(t, sc.Variable) else t.data
            if transformation_type == 'translation':
                v = v.to(unit='m', copy=False)
                v = sc.spatial.translations(dims=v.dims, values=v.values, unit=v.unit)
            elif transformation_type == 'rotation':
                v = sc.spatial.rotations_from_rotvecs(v)
            else:
                raise TransformationError(
                    f"{transformation_type=} attribute at {self.name},"
                    " expected 'translation' or 'rotation'."
                )
            if isinstance(t, sc.Variable):
                t = v
            else:
                t.data = v
            if (offset := self.offset) is None:
                return t
            offset = sc.vector(value=offset.values, unit=offset.unit).to(unit='m')
            offset = sc.spatial.translation(value=offset.value, unit=offset.unit)
            return t * offset
        except (sc.DimensionError, sc.UnitError) as e:
            raise NexusStructureError(
                f"Invalid transformation in NXtransformations: {e}"
            ) from e


def _interpolate_transform(transform, xnew):
    # scipy can't interpolate with a single value
    if transform.sizes["time"] == 1:
        transform = sc.concat([transform, transform], dim="time")
    return interpolate.interp1d(
        transform, "time", kind="previous", fill_value="extrapolate"
    )(xnew=xnew)


def _smaller_unit(a, b):
    if a.unit == b.unit:
        return a.unit
    ratio = sc.scalar(1.0, unit=a.unit).to(unit=b.unit)
    if ratio.value < 1.0:
        return a.unit
    else:
        return b.unit


def get_full_transformation(
    depends_on: Field,
) -> Union[None, sc.DataArray, sc.Variable]:
    """
    Get the 4x4 transformation matrix for a component, resulting
    from the full chain of transformations linked by "depends_on"
    attributes
    """
    if (t0 := make_transformation(depends_on, depends_on[()])) is None:
        return None
    return get_full_transformation_starting_at(t0)


def get_full_transformation_starting_at(
    t0: Transformation, *, index: ScippIndex = None
) -> Union[None, sc.DataArray, sc.Variable]:
    transformations = _get_transformations(t0, index=() if index is None else index)

    total_transform = None
    for transform in transformations:
        if total_transform is None:
            total_transform = transform
        elif isinstance(total_transform, sc.DataArray) and isinstance(
            transform, sc.DataArray
        ):
            unit = _smaller_unit(
                transform.coords['time'], total_transform.coords['time']
            )
            total_transform.coords['time'] = total_transform.coords['time'].to(
                unit=unit, copy=False
            )
            transform.coords['time'] = transform.coords['time'].to(
                unit=unit, copy=False
            )
            time = sc.concat(
                [total_transform.coords["time"], transform.coords["time"]], dim="time"
            )
            time = sc.datetimes(values=np.unique(time.values), dims=["time"], unit=unit)
            total_transform = _interpolate_transform(
                transform, time
            ) * _interpolate_transform(total_transform, time)
        else:
            total_transform = transform * total_transform
    if isinstance(total_transform, sc.DataArray):
        time_dependent = [t for t in transformations if isinstance(t, sc.DataArray)]
        times = [da.coords['time'][0] for da in time_dependent]
        latest_log_start = sc.reduce(times).max()
        return total_transform['time', latest_log_start:].copy()
    return total_transform


def _get_transformations(
    transform: Transformation, *, index: ScippIndex
) -> List[Union[sc.DataArray, sc.Variable]]:
    """Get all transformations in the depends_on chain."""
    transformations = []
    t = transform
    while t is not None:
        transformations.append(t[index])
        t = t.depends_on
    # TODO: this list of transformation should probably be cached in the future
    # to deal with changing beamline components (e.g. pixel positions) during a
    # live data stream (see https://github.com/scipp/scippneutron/issues/76).
    return transformations



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
# @author Simon Heybrock
from typing import List, Union

import numpy as np
import scipp as sc

from .._common import to_plain_index
from .nxobject import NexusStructureError, NXobject, ScippIndex

_event_dimension = "event"
_pulse_dimension = "pulse"


class NXevent_data(NXobject):
    @property
    def shape(self) -> List[int]:
        return self['event_index'].shape

    @property
    def dims(self) -> List[str]:
        return [_pulse_dimension]

    @property
    def unit(self) -> None:
        # Binned data, bins do not have a unit
        return None

    def _get_field_dims(self, name: str) -> Union[None, List[str]]:
        if name in ['event_time_zero', 'event_index']:
            return [_pulse_dimension]
        if name in ['event_time_offset', 'event_id']:
            return [_event_dimension]
        return None

    def _getitem(self, select: ScippIndex) -> sc.DataArray:
        self._check_for_missing_fields()
        index = to_plain_index([_pulse_dimension], select)

        max_index = self.shape[0]
        single = False
        if index is Ellipsis or index == tuple():
            last_loaded = False
        else:
            if isinstance(index, int):
                single = True
                start, stop, _ = slice(index, None).indices(max_index)
                if start == stop:
                    raise IndexError('Index {start} is out of range')
                index = slice(start, start + 1)
            start, stop, stride = index.indices(max_index)
            if stop + stride > max_index:
                last_loaded = False
            else:
                stop += stride
                last_loaded = True
            index = slice(start, stop, stride)

        event_index = self['event_index'][index].values
        event_time_zero = self['event_time_zero'][index]

        num_event = self["event_time_offset"].shape[0]
        # Some files contain uint64 "max" indices, which turn into negatives during
        # conversion to int64. This is a hack to get around this.
        event_index[event_index < 0] = num_event

        if len(event_index) > 0:
            event_select = slice(
                event_index[0], event_index[-1] if last_loaded else num_event
            )
        else:
            event_select = slice(None)

        if (event_id := self.get('event_id')) is not None:
            event_id = event_id[event_select]
            if event_id.dtype not in [sc.DType.int32, sc.DType.int64]:
                raise NexusStructureError(
                    "NXevent_data contains event_id field with non-integer values"
                )

        event_time_offset = self['event_time_offset'][event_select]

        # Weights are not stored in NeXus, so use 1s
        weights = sc.ones(
            dims=[_event_dimension],
            shape=event_time_offset.shape,
            unit='counts',
            dtype=np.float32,
        )

        events = sc.DataArray(
            data=weights, coords={'event_time_offset': event_time_offset}
        )
        if event_id is not None:
            events.coords['event_id'] = event_id

        if not last_loaded:
            event_index = np.append(event_index, num_event)
        else:
            # Not a bin-edge coord, all events in bin are associated with same
            # (previous) pulse time value
            # Copy to avoid confusing size display in _repr_html_
            event_time_zero = event_time_zero[:-1].copy()

        event_index = sc.array(
            dims=[_pulse_dimension], values=event_index, dtype=sc.DType.int64, unit=None
        )

        event_index -= event_index.min()

        # There is some variation in the last recorded event_index in files from
        # different institutions. We try to make sure here that it is what would be the
        # first index of the next pulse. In other words, ensure that event_index
        # includes the bin edge for the last pulse.
        if single:
            begins = event_index[_pulse_dimension, 0]
            ends = event_index[_pulse_dimension, 1]
            event_time_zero = event_time_zero[_pulse_dimension, 0].copy()
        else:
            begins = event_index[_pulse_dimension, :-1]
            ends = event_index[_pulse_dimension, 1:]

        try:
            binned = sc.bins(data=events, dim=_event_dimension, begin=begins, end=ends)
        except IndexError as e:
            raise NexusStructureError(
                f"Invalid index in NXevent_data at {self.name}/event_index:\n{e}."
            )

        return sc.DataArray(data=binned, coords={'event_time_zero': event_time_zero})

    def _check_for_missing_fields(self):
        for field in ("event_time_zero", "event_index", "event_time_offset"):
            if field not in self:
                raise NexusStructureError(
                    f"Required field {field} not found in NXevent_data"
                )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
# @author Simon Heybrock
from .base import NXobject, NXroot  # noqa F401
from .nxcylindrical_geometry import NXcylindrical_geometry  # noqa F401
from .nxdata import NXdata, NXdetector, NXlog, NXmonitor  # noqa F401
from .nxevent_data import NXevent_data  # noqa F401
from .nxoff_geometry import NXoff_geometry  # noqa F401
from .nxsample import NXsample  # noqa F401
from .nxtransformations import NXtransformations  # noqa F401


class NXentry(NXobject):
    """Entry in a NeXus file."""


class NXinstrument(NXobject):
    """Group of instrument-related information."""


class NXaperture(NXobject):
    """NXaperture"""


class NXattenuator(NXobject):
    """NXattenuator"""


class NXbeam(NXobject):
    """NXbeam"""


class NXbeam_stop(NXobject):
    """NXbeam_stop"""


class NXbending_magnet(NXobject):
    """NXbending_magnet"""


class NXcapillary(NXobject):
    """NXcapillary"""


class NXcite(NXobject):
    """NXcite"""


class NXcollection(NXobject):
    """NXcollection"""


class NXcollimator(NXobject):
    """NXcollimator"""


class NXcrystal(NXobject):
    """NXcrystal"""


class NXdetector_group(NXobject):
    """NXdetector_group"""


class NXdetector_module(NXobject):
    """NXdetector_module"""


class NXdisk_chopper(NXobject):
    """NXdisk_chopper"""


class NXenvironment(NXobject):
    """NXenvironment"""


class NXfermi_chopper(NXobject):
    """NXfermi_chopper"""


class NXfilter(NXobject):
    """NXfilter"""


class NXflipper(NXobject):
    """NXflipper"""


class NXfresnel_zone_plate(NXobject):
    """NXfresnel_zone_plate"""


class NXgeometry(NXobject):
    """NXgeometry"""


class NXgrating(NXobject):
    """NXgrating"""


class NXguide(NXobject):
    """NXguide"""


class NXinsertion_device(NXobject):
    """NXinsertion_device"""


class NXmirror(NXobject):
    """NXmirror"""


class NXmoderator(NXobject):
    """NXmoderator"""


class NXmonochromator(NXobject):
    """NXmonochromator"""


class NXnote(NXobject):
    """NXnote"""


class NXorientation(NXobject):
    """NXorientation"""


class NXparameters(NXobject):
    """NXparameters"""


class NXpdb(NXobject):
    """NXpdb"""


class NXpinhole(NXobject):
    """NXpinhole"""


class NXpolarizer(NXobject):
    """NXpolarizer"""


class NXpositioner(NXobject):
    """NXpositioner"""


class NXprocess(NXobject):
    """NXprocess"""


class NXreflections(NXobject):
    """NXreflections"""


class NXsample_component(NXobject):
    """NXsample_component"""


class NXsensor(NXobject):
    """NXsensor"""


class NXshape(NXobject):
    """NXshape"""


class NXsource(NXobject):
    """NXsource"""


class NXslit(NXobject):
    """NXslit"""


class NXsubentry(NXobject):
    """NXsubentry"""


class NXtranslation(NXobject):
    """NXtranslation"""


class NXuser(NXobject):
    """NXuser"""


class NXvelocity_selector(NXobject):
    """NXvelocity_selector"""


class NXxraylens(NXobject):
    """NXxraylens"""


# Not included in list of NeXus classes since this is the "base" of all others
del NXobject



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
# @author Simon Heybrock
from typing import Literal, Optional, Tuple, Union

import scipp as sc

from ..definition import ApplicationDefinition as BaseDef
from ..nxobject import NXobject


class ApplicationDefinition(BaseDef):
    def __init__(self, class_attribute: str, default: str = None):
        self._default_class = default
        self._class_attribute = class_attribute
        self._strategies = {}

    def make_strategy(self, group: NXobject):
        # This approach will likely need to be generalized as many application
        # definitions to not define a "class attribute" in the style of canSAS_class,
        # but seem to rely on basic structure and the NX_class attribute.
        if (
            definition_class := group.attrs.get(
                self._class_attribute, self._default_class
            )
        ) is not None:
            return self._strategies.get(definition_class)

    def register(self, sas_class):
        def decorator(strategy):
            self._strategies[sas_class] = strategy
            return strategy

        return decorator


NXcanSAS = ApplicationDefinition('canSAS_class', 'SASroot')


class SASdata:
    nx_class = 'NXdata'

    def __init__(
        self,
        data: sc.DataArray,
        Q_variances: Optional[Literal['uncertainties', 'resolutions']] = None,
    ):
        self.data = data
        valid = ('uncertainties', 'resolutions')
        if Q_variances not in (None,) + valid:
            raise ValueError(f"Q_variances must be in {valid}")
        self._variances = Q_variances

    def __write_to_nexus_group__(self, group: NXobject):
        da = self.data
        group.attrs['canSAS_class'] = 'SASdata'
        group.attrs['signal'] = 'I'
        group.attrs['axes'] = da.dims  # for NeXus compliance, same as I_axes
        group.attrs['I_axes'] = da.dims
        group.attrs['Q_indices'] = tuple(da.dims.index(d) for d in da.coords['Q'].dims)
        signal = group.create_field('I', sc.values(da.data))
        # We use the _errors suffix for NeXus compliance, unlike the examples given in
        # NXcanSAS.
        if da.variances is not None:
            signal.attrs['uncertainties'] = 'I_errors'
            group.create_field('I_errors', sc.stddevs(da.data))
        if da.coords.is_edges('Q'):
            raise ValueError(
                "Q is given as bin-edges, but NXcanSAS requires Q points (such as "
                "bin centers)."
            )
        coord = group.create_field('Q', da.coords['Q'])
        if da.coords['Q'].variances is not None:
            if self._variances is None:
                raise ValueError(
                    "Q has variances, must specify whether these represent "
                    "'uncertainties' or 'resolutions' using the 'Q_variances' option'"
                )

            coord.attrs[self._variances] = 'Q_errors'
            group.create_field('Q_errors', sc.stddevs(da.coords['Q']))


@NXcanSAS.register('SASdata')
class SASdataStrategy:
    @staticmethod
    def axes(group: NXobject) -> Tuple[str]:
        return group.attrs.get('I_axes')

    @staticmethod
    def signal(group: NXobject) -> str:
        return group.attrs.get('signal', 'I')

    @staticmethod
    def signal_errors(group: NXobject) -> Optional[str]:
        signal_name = group.attrs.get('signal', 'I')
        signal = group._group[signal_name]
        return signal.attrs.get('uncertainties')

    def coord_errors(group: NXobject, name: str) -> Optional[str]:
        if name != 'Q':
            return None
        # TODO This naively stores this as Scipp errors, which are just Gaussian.
        # This is probably not correct in all cases.
        uncertainties = group[name].attrs.get('uncertainties')
        resolutions = group[name].attrs.get('resolutions')
        if uncertainties is None:
            return resolutions
        elif resolutions is None:
            return uncertainties
        raise RuntimeError("Cannot handle both uncertainties and resolutions for Q")


@NXcanSAS.register('SAStransmission_spectrum')
class SAStransmission_spectrumStrategy:
    @staticmethod
    def dims(group: NXobject) -> Tuple[str]:
        # TODO A valid file should have T_axes, do we need to fallback?
        if (axes := group.attrs.get('T_axes')) is not None:
            return (axes,)
        return ('lambda',)


class SASentry:
    nx_class = 'NXentry'

    def __init__(self, *, title: str, run: Union[str, int]):
        self.title = title
        self.run = run

    def __write_to_nexus_group__(self, group: NXobject):
        group.attrs['canSAS_class'] = 'SASentry'
        group.attrs['version'] = '1.0'
        group.attrs['definition'] = 'NXcanSAS'
        group.create_field('title', self.title)
        group.create_field('run', self.run)


@NXcanSAS.register('SASentry')
class SASentryStrategy:
    pass


@NXcanSAS.register('SASroot')
class SASrootStrategy:
    pass



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
# @author Simon Heybrock

from typing import List, Union

import scipp as sc

from .nxdata import NXdata, NXdataStrategy
from .nxobject import NXobject, ScippIndex


class NXlogStrategy(NXdataStrategy):
    @staticmethod
    def axes(group):
        if (ax := NXdataStrategy.axes(group)) is not None:
            return ax
        # We get the shape from the original dataset, to make sure we do not squeeze
        # dimensions too early
        child_dataset = group._get_child('value')._dataset
        ndim = child_dataset.ndim
        shape = child_dataset.shape
        # The outermost axis in NXlog is pre-defined to 'time' (if present). Note
        # that this may be overridden by an `axes` attribute, if defined for the group.
        if 'time' in group:
            raw_axes = ['time'] + (['.'] * (ndim - 1))
        else:
            raw_axes = ['.'] * ndim
        axes = []
        for i, ax in enumerate(raw_axes):
            # Squeeze dimensions that have size 1 and are not 'time'
            if (ax == 'time') or (shape[i] != 1):
                axes.append(ax)
        return axes

    @staticmethod
    def signal(group):
        # NXdata uses the 'signal' attribute to define the field name of the signal.
        # NXlog uses a "hard-coded" signal name 'value', without specifying the
        # attribute in the file, so we pass this explicitly to NXdata.
        return group.attrs.get('signal', 'value')


class NXlog(NXobject):
    @property
    def shape(self):
        return self._nxbase.shape

    @property
    def dims(self):
        return self._nxbase.dims

    @property
    def unit(self):
        return self._nxbase.unit

    @property
    def _nxbase(self) -> NXdata:
        return NXdata(
            self._group,
            strategy=NXlogStrategy,
            skip=['cue_timestamp_zero', 'cue_index'],
        )

    def _getitem(self, select: ScippIndex) -> sc.DataArray:
        base = self._nxbase
        # Field loads datetime offset attributes automatically, but for NXlog this
        # may apparently be omitted and must then interpreted as relative to epoch.
        base.child_params['time'] = {'is_time': True}
        return base[select]

    def _get_field_dims(self, name: str) -> Union[None, List[str]]:
        return self._nxbase._get_field_dims(name)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
# @author Simon Heybrock

# flake8: noqa

from .nxcansas import SASdata, SASentry, definitions

__all__ = ['definitions', 'SASentry', 'SASdata']
